from ridt.base import ComputationalSpace

from ridt.config import RIDTConfig

from .batchdatastore import BatchDataStore
from .directoryagent import DirectoryAgent
from .datastoreplotter import DataStorePlotter


class BatchDataStorePlotter:
    """A class which handles the plotting of a batch data store.

    Iterates through a :class:`~.BatchDataStore` instance and runs  the
    :class:`~.DataStorePlotter` on them.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.
    
    outdir: :obj:`str`
        The path to the output directory for the run.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    data_store : :class:`~.BatchDataStore`
        The batch data store to be analysed.

    """
    def __init__(self,
                 settings: RIDTConfig,
                 data_store: BatchDataStore,
                 space: ComputationalSpace,
                 outdir: str,
                 quantity: str):
        """The :class:`~.BatchDataStorePlotter` class initialiser.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        data_store : :class:`~.BatchDataStore`
            The batch data store to be analysed.

        space : :class:`~.ComputationalSpace`
            The :class:`~.ComputationalSpace` instance corresponding to the
            :attr:`settings` attribute.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.settings = settings
        self.data_store = data_store
        self.outdir = outdir
        self.space = space
        self.quantity = quantity
        print(f"Plotting {self.quantity} data...")
        self.plot()

    def plot(self) -> None:
        """The method that loops through the :class:`~.DataStore` instances

        and calls the :class:`~.DataStorePlotter` class on them. This method
        will create the relevant subdirectories for output if they do not
        already exist.

        Returns
        -------
        None

        """
        dir_agent = DirectoryAgent(self.outdir, self.space.shape)
        arg = lambda x: (dir_agent, self.data_store[x], x, self.quantity)

        if self.space.zero:
            DataStorePlotter(*arg(self.settings))
        else:
            for idx, setting in enumerate(self.space.space):
                count = f"{idx + 1}/{len(self.space)}"
                print(f"Plotting computational space element {count}")
                dir_agent.create_root_dir(idx)
                DataStorePlotter(*arg(setting))


import unittest
import json
import os

import numpy as np

from ridt.config import RIDTConfig
from ridt.equation import EddyDiffusion


class TestEddyDiffusion(unittest.TestCase):

    def setUp(self) -> None:


        this_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(this_dir, "test_resources/test_config.json")) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)

        self.ed = EddyDiffusion(self.config)

        self.time_array = np.linspace(0, 10, 10)
        self.x_space = np.linspace(0, 10, 10)
        self.y_space = np.linspace(0, 10, 10)
        self.z_space = np.linspace(0, 10, 10)

    def test_outputs(self):

        """Ensures the concentration is of type array"""
        X, Y, Z = np.meshgrid(self.x_space,
                              self.y_space,
                              self.z_space)

        concentration = self.ed(X, Y, Z, self.time_array)
        self.assertEqual(type(concentration), np.ndarray)


if __name__ == "__main__":
    unittest.main()



import unittest
import shutil

from os.path import join
from os.path import dirname
from os.path import isfile
from os.path import isdir
from os.path import islink
from os import listdir
from os import system
from os import unlink

from numpy import genfromtxt
from numpy import load
from numpy import delete
from numpy import ndarray
from numpy import array
from numpy import abs

from ridt.config import ConfigFileParser

from ridt.config.ridtconfig import FixedDurationSource
from ridt.config.ridtconfig import InstantaneousSource
from ridt.config.ridtconfig import InfiniteDurationSource

def delete_folder_contents(path: str):
    for filename in listdir(path):
        file_path = join(path, filename)
        try:
            if isfile(file_path) or islink(file_path):
                unlink(file_path)
            elif isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))

class ST27(unittest.TestCase):
    """System Test 28. Check output against verification data."""
    def conf(self, cat: str):
        return join(self.root, cat, "config.json")
    
    def out(self, cat: str):
        return join(self.root, cat, "out")
    
    def points(self, cat: str):
        path = join(self.root, cat, "points.csv")
        with open(path, 'r', encoding='utf-8-sig') as f:
            return genfromtxt(f, delimiter=',')
    
    def lines(self, cat: str):
        path = join(self.root, cat, "lines.csv")
        with open(path, 'r', encoding='utf-8-sig') as f:
            return genfromtxt(f, delimiter=',')
    
    def percentage_error(self, test_data: ndarray, val_data: ndarray):
        return abs(((test_data - val_data) / val_data)) * 100

    def loadpoint(self, cat: str):
        path = join(
            self.root,
            cat,
            "out",
            "points",
            "concentration",
            "data",
            f"well_mixed.npy")
        return load(path)

    def setUp(self) -> None:
        self.root = join(dirname(__file__), "st29")
        self.cat = ["ins", "inf", "fix"]

    def tearDown(self) -> None:
        for cat in self.cat:
            delete_folder_contents(self.out(cat))

    def test_ins(self):
        idstr = "ins"
        print("")
        print(idstr)

        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        point = self.loadpoint(idstr)

        for idx, item in enumerate(self.percentage_error(point, valpoints)):
            self.assertTrue(item < 0.1)
        print("")

    def test_inf(self):
        idstr = "inf"
        print("")
        print(idstr)

        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        point = self.loadpoint(idstr)
        point = delete(point, 0)
        valpoints = delete(valpoints, 0)

        for idx, item in enumerate(self.percentage_error(point, valpoints)):
            self.assertTrue(item < 0.1)
        print("")

    def test_fix(self):
        idstr = "fix"
        print("")
        print(idstr)

        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        point = self.loadpoint(idstr)
        point = delete(point, 0)
        valpoints = delete(valpoints, 0)

        for idx, item in enumerate(self.percentage_error(point, valpoints)):
            self.assertTrue(item < 0.1)
        print("")




if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.config import ConfigFileParser


class ST10(unittest.TestCase):

    """System Test 10. Test the system can
       parse the user’s choice of diffusion
       coefficient from the JSON file."""

    def setUp(self) -> None:

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "coefficient"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.coefficient, "calculation"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.coefficient, "value"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.coefficient, "tkeb"), True
        )
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.coefficient.tkeb, "number_of_supply_vents"), True
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import os

from ridt.config import ConfigFileParser


class ST04(unittest.TestCase):

    """System Test 04. Test that the system is
       able to parse all physical parameters
       defined in the user requirements from
       the JSON file."""

    def setUp(self) -> None:

        """setUp method which instantiates an
        :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def test_verify(self):

        """Verify that the :class:`~.RIDTConfig` class contains
        the correct attribute."""

        self.assertEqual(
            hasattr(self.c, "eddy_diffusion"), True)
        self.assertEqual(
            hasattr(self.c, "well_mixed"), True)
        self.assertEqual(
            hasattr(self.c, "time_samples"), True)
        self.assertEqual(
            hasattr(self.c, "total_time"), True)
        self.assertEqual(
            hasattr(self.c, "fresh_air_flow_rate_units"), True)
        self.assertEqual(
            hasattr(self.c, "fresh_air_flow_rate"), True)
        self.assertEqual(
            hasattr(self.c, "modes"), True)
        self.assertEqual(
            hasattr(self.c, "models"), True)


if __name__ == "__main__":
    unittest.main()



import unittest

import os

import json

from ridt.config import ConfigFileParser
from ridt.config.ridtconfig import ConsistencyError
from ridt.config.configfileparser import ConfigFileParserValidationError


class ST02(unittest.TestCase):

    """System Test 02. The objective is to able to parse monitor
       points/lines/planes parallel to the principle axes from the JSON file,
       to be evaluated by the Eddy Diffusion model."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.config_path = os.path.join(this_dir, "st06/config.json")
        with open(self.config_path) as f:
            self.default = json.load(f)

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def tearDown(self) -> None:

        with open(self.config_path, "w") as w:
            json.dump(self.default, w)

    def test_verify(self):

        """"Verifies that the :class:`~.RIDTConfig` class contains the
        monitor locations attributes."""

        self.assertEqual(
            hasattr(self.c, "models"), True)
        self.assertEqual(
            hasattr(self.c.models, "eddy_diffusion"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion, "monitor_locations"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.monitor_locations, "points"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.monitor_locations, "lines"), True)
        self.assertEqual(
            hasattr(self.c.models.eddy_diffusion.monitor_locations, "planes"), True)

    def test_verify_points(self):

        """Verifies that the :class:`~.Point` class contains the
        correct attributes."""

        for point in self.c.models.eddy_diffusion.monitor_locations.points.values():
            self.assertEqual(hasattr(point, "x"), True)
            self.assertEqual(hasattr(point, "y"), True)
            self.assertEqual(hasattr(point, "z"), True)

    def test_verify_lines(self):

        """Verifies that the :class:`~.Line` class contains the
        correct attributes."""

        for line in self.c.models.eddy_diffusion.monitor_locations.lines.values():
            self.assertEqual(hasattr(line, "point"), True)
            self.assertEqual(hasattr(line.point, "x"), True)
            self.assertEqual(hasattr(line.point, "y"), True)
            self.assertEqual(hasattr(line.point, "z"), True)
            self.assertEqual(hasattr(line, "parallel_axis"), True)

    def test_verify_planes(self):

        """Verifies that the :class:`~.Plane` class contains the
        correct attributes."""

        for plane in self.c.models.eddy_diffusion.monitor_locations.planes.values():
            self.assertEqual(hasattr(plane, "axis"), True)
            self.assertEqual(hasattr(plane, "distance"), True)

    def test_line_dims(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers if the line does not lie in the bounds of the container
        or is not parallel to one of the principle axes."""

        with self.assertRaises(ConsistencyError):
            with open(self.config_path) as f:
                config = json.load(f)
            for lines in config["models"]["eddy_diffusion"]["monitor_locations"]["lines"].values():
                lines["point"]["x"] = config["dimensions"]["x"] + 1
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_line_axis(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers if the line does not lie in the bounds of the container
        or is not parallel to one of the principle axes."""

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            for lines in config["models"]["eddy_diffusion"]["monitor_locations"]["lines"].values():
                lines["parallel_axis"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_points(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers if the point does not lie in the bounds of the container."""

        with self.assertRaises(ConsistencyError):
            with open(self.config_path) as f:
                config = json.load(f)
            for points in config["models"]["eddy_diffusion"]["monitor_locations"]["points"].values():
                points["x"] = config["dimensions"]["x"] + 1
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_planes_axis(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers if a plane lies outside of the bounds of the container."""

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            for planes in config["models"]["eddy_diffusion"]["monitor_locations"]["planes"].values():
                planes["axis"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_planes_distance(self):

        """Checks to see if the :class:`.ConsistencyError` error
        triggers if a plane lies outside of the bounds of the container."""

        with self.assertRaises(ConsistencyError):
            with open(self.config_path) as f:
                config = json.load(f)
            for planes in config["models"]["eddy_diffusion"]["monitor_locations"]["planes"].values():
                axis = [x for x in ["x", "y", "z"] if x not in list(planes["axis"])]
                planes["distance"] = config["dimensions"][axis[0]] + 1
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)


if __name__ == "__main__":
    unittest.main()



from .settings import Settings
from .settings import Dict
from .settings import List
from .settings import Terminus
from .settings import Number
from .settings import StringSelection
from .settings import ComputationalSpace
from .settings import SettingErrorMessage
from .settings import ConsistencyError

from .exceptions import Error
from .exceptions import ListTypeError
from .exceptions import FloatTypeError
from .exceptions import IntTypeError
from .exceptions import DimensionError
from .exceptions import RIDTOSError



import unittest
import json
import os

from ridt.data import BatchDataStore
from ridt.data import DataStore

from ridt.config import RIDTConfig


class TestBatchDataStore(unittest.TestCase):

    "Unit tests for the :class:`~.BatchDataStore` class."""

    def setUp(self) -> None:

        """setUp method which instantiates the :class:`~.RIDTConfig` class,
        and the :class:`~.BatchDataStore` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with open(path) as f:
            loaded_json = json.load(f)

        self.config = RIDTConfig(loaded_json)

        self.bds = BatchDataStore()

    def test_add_run(self):

        """Adds a run to the batch data store then checks
        to make sure the new items match the type of
        the ones added."""

        self.bds.add_run(self.config)
        self.assertEqual(type(self.bds.store), dict)
        for key, val in self.bds.store.items():
            self.assertEqual(type(key), RIDTConfig)
            self.assertEqual(type(val), DataStore)



from typing import Union

from copy import deepcopy

from numpy import array
from numpy import ndarray
from numpy.linalg import norm
from numpy import nan

from ridt.config import RIDTConfig
from ridt.config import InstantaneousSource
from ridt.config import InfiniteDurationSource
from ridt.config import FixedDurationSource 

from ridt.container import Domain

Source = Union[InstantaneousSource, InfiniteDurationSource, FixedDurationSource]


class UncertaintyMask:
    """Provides the ability to mask uncertain values in quantity grids.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run.

    sources : :obj:`List`[:class:`~numpy.ndarray`]
        The list of source term positions found in the settings instance.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`settings`.
    
    radius : :obj:`float`
        The radius of the sphere around each source, inside which values will be
        masked.

    """
    def __init__(self, setting: RIDTConfig):
        """The :class:`~.UncertaintyMask` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run.

        """
        self.setting = setting
        self.radius = setting.models.eddy_diffusion.analysis.exclude_radius_meters
        self.sources = self.get_source_locations()
        self.domain = Domain(setting)

    def get_source_locations(self):
        """Get the source settings objects from the run settings instance.

        Returns
        -------
        :obj:`List`[:class:`~numpy.ndarray`]
            The list of source term positions found in the settings instance.

        """
        rv = list()
        modes = ["instantaneous", "infinite_duration", "fixed_duration"]
        for mode in modes:
            sources = getattr(self.setting.modes, mode).sources
            rv += [array((s.x, s.y, s.z)) for s in sources.values()]
        return rv

    def mask(self, geometry: str, id: str, data: ndarray):
        """Masks all elements in the grid that are within 2m of a source.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be masked.

        id : :obj:`str`
            The id of the grid to be masked.

        data : :class:`~ndarray.ndarray`
            The grid to be masked.

        Returns
        -------
        :class:`~numpy.ndarray`
            The masked array.

        """
        print(f"Excluding uncertain values in {id}...")
        return getattr(self, f"mask_{geometry}")(id, data)
    
    def mask_points(self, id: str, data: ndarray):
        """Mask values in point array close to a source.

        Parameters
        ----------
        id : :obj:`str`
            The id of the point.

        data : :class:`~.numpy.ndarray`
            The array to be masked.

        Returns
        -------
        :class:`~.numpy.ndarray`
            The masked copy.

        """
        point = self.setting.models.eddy_diffusion.monitor_locations.points[id]
        location = self.domain.point_cartesian(point)
        rv = deepcopy(data)
        for s in self.sources:
            if norm(s - location) <= self.radius:
                rv.fill(nan)
        return rv
        
    def mask_lines(self, id: str, data: ndarray):
        """Mask values in line array close to a source.

        Parameters
        ----------
        id : :obj:`str`
            The id of the line.

        data : :class:`~.numpy.ndarray`
            The array to be masked.

        Returns
        -------
        :class:`~.numpy.ndarray`
            The masked copy.

        """
        line = self.setting.models.eddy_diffusion.monitor_locations.lines[id]
        locations = self.domain.line_cartesian(line)
        rv = deepcopy(data)
        for idx, location in locations:
            for s in self.sources:
                if norm(s - array(location)) <= self.radius:
                    for idt, time in enumerate(self.domain.time):
                        rv[idt][idx] = nan
        return rv
        
    def mask_planes(self, id: str, data: ndarray):
        """Mask values in plane array close to a source.

        Parameters
        ----------
        id : :obj:`str`
            The id of the plane.

        data : :class:`~.numpy.ndarray`
            The array to be masked.

        Returns
        -------
        :class:`~.numpy.ndarray`
            The masked copy.

        """
        plane = self.setting.models.eddy_diffusion.monitor_locations.planes[id]
        locations = self.domain.plane_cartesian(plane)
        rv = deepcopy(data)
        for idx, location in locations:
            for s in self.sources:
                if norm(s - array(location)) <= self.radius:
                    for idt, time in enumerate(self.domain.time):
                        rv[idt][idx] = nan
        return rv
        
    def mask_domain(self, id: str, data: ndarray):
        """Mask values in domain array close to a source.

        Note
        ----
        The `id` parameter is a placeholder for future functionality, where
        multiple domains can be defined.

        Parameters
        ----------
        id : :obj:`str`
            The id of the domain.

        data : :class:`~.numpy.ndarray`
            The array to be masked.

        Returns
        -------
        :class:`~.numpy.ndarray`
            The masked copy.

        """
        locations = self.domain.domain_cartesian()
        rv = deepcopy(data)
        for idx, location in locations:
            for s in self.sources:
                if norm(s - array(location)) <= self.radius:
                    for idt, time in enumerate(self.domain.time):
                        rv[idt][idx] = nan
        return rv



from .ridtconfig import RIDTConfig
from .ridtconfig import Point
from .ridtconfig import Line
from .ridtconfig import Plane
from .ridtconfig import InstantaneousSource
from .ridtconfig import InfiniteDurationSource 
from .ridtconfig import FixedDurationSource 

from .configfileparser import ConfigFileParser
from .configfileparser import ConfigFileParserJSONError

from .configfilewriter import ConfigFileWriter

from .csvtoconfigfile import CSVToConfigFile

from .summary import summary

from .units import Units



import unittest
import json
import os

import numpy as np

from ridt.config import ConfigFileWriter
from ridt.config.configfilewriter import ConfigFileWriterOSError


class TestConfigFileWriter(unittest.TestCase):

    """Unit tests for the :class:`~.ConfigFileWriter` class."""

    def setUp(self) -> None:

        """setUp method which creates a dummy data :obj:`dict`
        which is used within the tests. Additionally it
        instantiates the :class:`~.ConfigFileWriter` class."""

        self.data = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}

    def test_write(self):

        """Tests the write method to ensure that what is written
        into the json matches the given :obj:`dict`"""

        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        ConfigFileWriter(
            os.path.join(self.this_dir, "test_resources"),
            "configfilewritertest.json",
            self.data)

        with open(os.path.join(self.this_dir, "test_resources/configfilewritertest.json")) as f:
            test_file = json.load(f)
        self.assertEqual(test_file, self.data)

    def test_OSError(self):

        """Ensures that the :class:`~.ConfigFileWriterOSError`
        passes when the file path is incorrect."""

        with self.assertRaises(ConfigFileWriterOSError):
            ConfigFileWriter(
                "tests/test_resource",
                "testfilepath.json",
                self.data)



import warnings

from os.path import join

from numpy import ndarray

from ridt.config import RIDTConfig
from ridt.config import Units

from ridt.container import Domain

import matplotlib.pyplot as plt


class LinePlot:
    """The class that generates line plots over some spatial domain.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.
    
    config : :class:`~.LinePlots`
        The line plots settings object.

    outdir: :obj:`str`
        The path to the output directory for the run.

    """

    def __init__(self, settings: RIDTConfig, output_dir: str, quantity: str):
        """The :class:`~.LinePlot` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.settings = settings
        self.output_dir = output_dir
        self.units = Units(settings)
        self.domain = Domain(self.settings)
        self.quantity = quantity
        self.config = self.settings.models.eddy_diffusion.lines_plots

    def __call__(self, id: str, data: ndarray, max_val: float, t_index: int) -> None:
        """The call method used to plot some data.

        Parameters
        ----------
        id : :obj:`str`
            The id string of the plane being plotted.

        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        max_val : :obj:`float`
            The maximum value over the whole time domain.

        t_index : :obj:`int`
            The time index of the current data array.

        """
        self.id = id
        self.t_index = t_index
        self.max_val = max_val
        self.axis = self.get_axis()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.plot(data)
            plt.tight_layout()
            self.save_fig()
            plt.close()

    def plot(self, data: ndarray):
        """Calls various plotting functions.

        Parameters
        ----------
        data : :class:`~numpy.ndarray`
            The array of values to be plot.

        """
        plt.title(self.title())
        plt.xlabel(self.xlabel())
        plt.ylabel(self.ylabel())

        if self.config.scale == "logarithmic":
            plt.yscale("log")

        plot = plt.plot(self.get_domain(), data, marker='.')
        return plot

    def save_fig(self):
        """Creates the formatted string file name and saves plot to disk.

        """
        plt.savefig(join(self.output_dir, f"{self.id}-{self.domain.time[self.t_index]:.2f}s.png"))

    def title(self):
        """Creates formatted title string for the current plot.

        Returns
        -------
        :obj:`str`
            The title string.

        """
        return f"{self.quantity} - {self.id} - {self.domain.time[self.t_index]:.2f}s"
    
    def ylabel(self):
        """Creates the formatted y-axis label string.

        Returns
        -------
        :obj:`str`
            The x-axis label string.

        """
        return f"{self.quantity} ({getattr(self.units, f'{self.quantity}')})"

    def xlabel(self):
        """Creates the formatted x-axis label string.

        Returns
        -------
        :obj:`str`
            The x-axis label string.

        """
        return f"{self.axis} ({self.units.space})"

    def get_domain(self):
        """Extract the x-axis values from the :attr:`domain`

        Returns
        -------
        :obj:`Iterable`[:obj:`float`]
            The x-axis values.

        """
        return getattr(self.domain, self.axis)

    def get_axis(self):
        """Parse the x axis from the plane settings object.

        Returns
        -------
        :obj:`str`
            The axis for the plot x, y, z  

        """
        lines = self.settings.models.eddy_diffusion.monitor_locations.lines
        return lines[self.id].parallel_axis



from typing import List

from numpy import ndarray
from numpy import zeros
from numpy import array
from numpy import finfo
from numpy import float64

from ridt.config import RIDTConfig

import numpy as np

FloatList = List[float]


class WellMixed:
    """The core Well Mixed model class.

    This class provides all equation implementation of the Well Mixed Model.

    For details of the mathematics implemented here, please see the user guide.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    dim : :class:`~.ridt.config.ridtconfig.Dimensions`
        The dimensions settings object that define the bounds of the system.

    volume : :obj:`float`
        The total volume of the system.

    fa_rate : :obj:`float`
        The fresh air change rate.

    shape : :obj:`Tuple`[:obj:`int`]
        The shape of the current grid.
    
    conc : :class:`~numpy.ndarray`
        The array where computed values are stored.
    
    """
    def __init__(self, settings: RIDTConfig):
        """The :class:`WellMixed` constructor.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.
            
        """
        self.settings = settings
        self.dim = self.settings.dimensions
        self.volume = self.dim.x * self.dim.y * self.dim.z
        self.fa_rate = settings.fresh_air_flow_rate
        self.fa_rate = self.fa_rate if self.fa_rate else finfo(float64).tiny
        self.shape = (self.settings.time_samples,)
        self.conc = zeros(self.shape)

    def __call__(self, t: FloatList):
        """This call method is used to evaluate the model.

        t : :obj:`List`[:obj:`float`]
            The current time domain array.

        Returns
        -------
        :class:`~numpy.ndarray`
            The calculated concentration values.

        """

        modes = ["instantaneous", "infinite_duration", "fixed_duration"]

        for mode in modes:
            self.sources = getattr(self.settings.modes, mode).sources
            getattr(self, f"{mode}")(t)
        return array(self.conc)

    def concentration(self, t: float):
        """The exponential decay equation.

        Parameters
        ----------
        t : :obj:`float`
            The time.

        Returns
        -------
        :obj:`float`
            The computed value.

        """
        return np.exp(-(self.fa_rate / self.volume) * t)

    def instantaneous(self, t: np.ndarray):
        """Evaluate all instanteneous sources at time `t`.

        Parameters
        ----------
        t : :obj:`float`
            The time at which to evaluate the model.

        Returns
        -------
        None

        """
        for idx, time in enumerate(t):
            for source in self.sources.values():
                if time - source.time >= 0:
                    self.conc[idx] += (source.mass / self.volume) *\
                        self.concentration(time - source.time)

    def infinite_duration(self, t: ndarray):
        """Evaluate all infinite duration sources at time `t`.

        Parameters
        ----------
        t : :obj:`float`
            The time at which to evaluate the model.

        Returns
        -------
        None

        """
        for idx, time in enumerate(t):
            for source in self.sources.values():
                if time - source.time >= 0:
                    self.conc[idx] += (source.rate / self.fa_rate) *\
                        (1 - self.concentration(time - source.time))

    def fixed_duration(self, t: ndarray):
        """Evaluate all infinite duration sources at time `t`.

        Parameters
        ----------
        t : :obj:`float`
            The time at which to evaluate the model.

        Returns
        -------
        None

        """
        for source in self.sources.values():
            end_int = 0
            temp_conc = zeros(self.shape)
            for idx, time in enumerate(t):
                if time < source.start_time:
                    pass
                elif time <= source.end_time:
                    temp_conc[idx] += (source.rate / self.fa_rate) *\
                        (1 - self.concentration(time - source.start_time))
                else:
                    if not end_int:
                        end_int = idx - 1
                    temp_conc[idx] += temp_conc[end_int] * self.concentration(
                        time - source.end_time)
            self.conc = [sum(x) for x in zip(self.conc, temp_conc)]



from ridt.base import ComputationalSpace

from ridt.config import RIDTConfig

from ridt.data import BatchDataStore
from ridt.data import DirectoryAgent
from ridt.data import DataStore

from .datastoreanalyser import DataStoreAnalyser
from .resultswriter import ResultsWriter
from .batchresultswriter import BatchResultsWriter

BF = '{l_bar}{bar:30}{r_bar}{bar:-10b}'


class BatchDataStoreAnalyser:
    """Batch data store analyser class.

    Iterates through a :class:`~.BatchDataStore` instance and runs  the
    :class:`~.DataStoreAnalyser` and :class:`~.ResultsWriter` classes on them.

    Attributes
    ----------
    settings : :class:`~.RIDTConfig`
        The settings for the run in question.

    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.
    
    outdir: :obj:`str`
        The path to the output directory for the run.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    data_store : :class:`~.BatchDataStore`
        The batch data store to be analysed.
    
    dir_agent: :class:`~.DirectoryAgent`
        The  :class:`~.DirectoryAgent` instance for this run.
    
    results : :obj:`dict` [:class:`~.RIDTConfig`, :class:`~.DataStoreAnalyser`]
        A dictionary of :class:`~.DataStoreAnalyser` instances for each
        :class:`~.RIDTConfig` object in :attr:`space`.

    """
    def __new__(cls, *args, **kwargs):
        instance = super(BatchDataStoreAnalyser, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return None

    def __init__(self,
                 settings: RIDTConfig,
                 data_store: BatchDataStore,
                 space: ComputationalSpace,
                 outdir: str,
                 quantity: str):
        """The :class:`~.BatchDataStoreAnalyser` class initialiser.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        data_store : :class:`~.BatchDataStore`
            The batch data store to be analysed.

        space : :class:`~.ComputationalSpace`
            The :class:`~.ComputationalSpace` instance corresponding to the
            :attr:`settings` attribute.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

        """
        self.settings = settings
        self.space = space
        self.outdir = outdir
        self.quantity = quantity
        self.dir_agent = DirectoryAgent(outdir, self.space.shape)
        self.data_store = data_store
        self.results = dict()
        print(f"Analysing {self.quantity}...")
        self.analyse()
    
    def analyse(self) -> None:
        """The method that loops through the :class:`~.DataStore` instances

        and calls the :class:`~.DataStoreAnalyser` and :class:`~.ResultsWriter`
        classes on them. This method will create the relevant subdirectories
        for output if they do not already exist.

        Returns
        -------
        None
    
        """
        if self.space.zero:
            store = self.data_store[self.settings]
            result = DataStoreAnalyser(self.settings, store, self.quantity)
            self.results[self.settings] = result
            ResultsWriter(self.settings, result, self.dir_agent, self.quantity)
        else:
            for setting, store in self.data_store.items():
                idx = self.space.linear_index(setting)
                print(f"Analysing computational space element {idx + 1}/{len(self.space)}")
                self.dir_agent.create_root_dir(idx)
                result = DataStoreAnalyser(setting, store, self.quantity)
                self.results[setting] = result
                ResultsWriter(setting, result, self.dir_agent, self.quantity)
            BatchResultsWriter(self.settings, self.space, self.results, self.outdir, self.quantity)



from typing import Tuple

from numpy import ndarray
from numpy import nanargmax
from numpy import unravel_index
from numpy import where
from numpy import prod
from numpy import isnan
from numpy import count_nonzero

from ridt.base import Error


class DataStore:
    """The data store than contains computed quantities over various geometries.

    There are four different sub stores for point, line, plane and full domain 
    grids. They can store quantities corresponding to different monitor
    locations defined in a settings file. The store data for various times (axis
    0) It also provides calculation of various quantities over the various
       grids.

    Attributes
    ----------
    points : :obj:`Dict`[:obj:`str`, :class:`~numpy.ndarray`]
        The dictionary containing all point monitor location value grids.

    lines : :obj:`Dict`[:obj:`str`, :class:`~numpy.ndarray`]
        The dictionary containing all line monitor location value grids.

    planes : :obj:`Dict`[:obj:`str`, :class:`~numpy.ndarray`]
        The dictionary containing all plane monitor location value grids.

    domain : :obj:`Dict`[:obj:`str`, :class:`~numpy.ndarray`]
        The dictionary containing all full domain value grids.

    """
    class Dimensions:
        """A 'static' member class that contains information about the

        dimensionality of the different geometry types.

        """
        points = 1
        lines = 2
        planes = 3
        domain = 4

    def __init__(self):
        """The :class:`DataStore` constructor.

        """
        self.points = dict()
        self.lines = dict()
        self.planes = dict()
        self.domain = dict()
    
    def add(self, geometry: str, id: str, data: ndarray)-> None:
        """Adds a new item to the data store.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be stored.

        id : :obj:`str`
            The id of the grid to be stored.

        data : :class:`~ndarray.ndarray`
            The grid to be stored.

        Raises
        ------
        :class:`~.DataStoreGeometryError`
            If the passed geometry type is invalid.

        Returns
        -------
        None

        """
        try:
            self.verify(data, getattr(DataStore.Dimensions, geometry))
            getattr(self, geometry)[id] = data
        except AttributeError:
            raise DataStoreGeometryError(geometry)
    
    def get(self, geometry: str, id: str) -> ndarray:
        """Returns the grid given a geometry type and id string.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be found.

        id : :obj:`str`
            The id of the grid to be found.

        Returns
        -------
        :class:`~numpy.ndarray`
            The grid to be stored.

        Raises
        ------
        :class:`~.DataStoreGeometryError`
            If the passed geometry type is invalid.
        :class:`~.DataStoreIDError`
            If the passed ID does not exist in the store..

        """
        try:
            return getattr(self, geometry)[id]
        except AttributeError:
            raise DataStoreGeometryError(geometry)
        except KeyError:
            raise DataStoreIDError(id, geometry)

    def verify(self, data: ndarray, dimensions: int) -> None:
        """Verifies if a grid is valid.

        Parameters
        ----------
        data : :class:`~ndarray.ndarray`
            The grid to be verified.

        dimensions : :obj:`int`
            The number of dimensions the grid should have.

        Raises
        ------
        :class:`~.DataStoreTypeError`
            If the grid is not a numpy array.

        :class:`~.DataStoreDimensionalityError`
            If the grid has the incorrect number of dimensions.
        """
        if not isinstance(data, ndarray):
            raise DataStoreTypeError(type(data))
        if len(data.shape) is not dimensions:
            raise DataStoreDimensionalityError(len(data.shape), dimensions)
    
    def maximum(self, geometry: str, id: str) -> Tuple[Tuple[int], ndarray]:
        """Finds and returns the maximum value, and its index, of a grid.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be evaluated.

        id : :obj:`str`
            The id of the grid to be evaluated.

        Returns
        -------
        :obj:`Tuple`[:obj:`Tuple`[:obj:`int`], float]
            The index and value of the maximum in the grid.

        """
        data = self.get(geometry, id)
        try:
            index = unravel_index(nanargmax(data), data.shape)
        except ValueError:
            index = None
        return index, data[index]
    
    def exceeds(self, geometry: str, id: str, value: float) -> Tuple[int]:
        """Returns the index of the first time a grid exceeds 'value'

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be evaluated.

        id : :obj:`str`
            The id of the grid to be evaluated.

        value : :obj:`float`
            The threshold to be exceeded.

        Returns
        -------
        :obj:`Union`[:obj:`Tuple`[:obj:`int`], None]
            The index where and when the first exceedence occurred, or None if
            no exceedence.

        """
        data = self.get(geometry, id)
        exceeds = self.zip(where(data >= value))
        return exceeds[0] if exceeds else None

    def percentage_exceeds(self, geometry: str, id: str, value: float, percent: float) -> int:
        """Returns the time index where 'percent'% of the grid exceeds 'value'

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be evaluated.

        id : :obj:`str`
            The id of the grid to be evaluated.

        value : :obj:`float`
            The threshold to be exceeded.

        percent : :obj:`float`
            The percentage of the grid required to exceed 'value'.

        Returns
        -------
        :obj:`Union`[:obj:`int`, None]
            The time index where the percentage exceedence was reached or None
            if not reached during simulation.

        """
        data = self.get(geometry, id)
        for time in range(data.shape[0]):
            size = data[time].size - count_nonzero(isnan(data[time]))
            if not size:
                continue
            frac = 100 * len(self.zip(where(data[time] >= value))) / size
            if frac >= percent:
                return time
        return None
    
    def percentage_exceeds_max(self, geometry: str, id: str, value: float) -> Tuple[int, float]:
        """Returns the maximum percentage of grid that exceeds `value`.

        And the time index when it occurred.

        Parameters
        ----------
        geometry : :obj:`str`
            The type of grid to be evaluated.

        id : :obj:`str`
            The id of the grid to be evaluated.

        value : :obj:`float`
            The threshold to be exceeded.

        Returns
        -------
        :obj:`Tuple`[:obj:`Union`[:obj:`int`, None], :obj:`float`]
            The time index and the percentage exceedence, or None and zero if no
            exceedence.

        """
        data = self.get(geometry, id)
        max_val = 0.0
        max_time = None
        shape = data.shape

        for time in range(shape[0]):
            size = data[time].size - count_nonzero(isnan(data[time]))
            if not size:
                continue
            frac = 100 * len(self.zip(where(data[time] >= value))) / size
            if frac > max_val:
                max_val = frac
                max_time = time
        return max_time, max_val
    
    def zip(self, where_output):
        return list(zip(*where_output))
 

class DataStoreIDError(Error):
    """The exception raised when the data store is queries with an id string
    it does not recognise.

    """
    def __init__(self, id_str: str, kind: str):
        """The constructor for the :class:`DataStoreIDError` class.

        """
        if id_str:
            msg = f"The data store does not contain any {kind} data with ID "\
                  f"{id_str}."
        else:
            msg = f"The data store does not contain any {kind} data."
        super().__init__(msg)


class DataStoreDimensionalityError(Error):
    """The exception raised when an array of the wrong shape is passed to the
    data store.

    """
    def __init__(self, dimensions: int, expected_dimensions: int):
        """The constructor for the :class:`DataStoreDimensionalityError` class.

        """
        msg = f"Expecting array with {expected_dimensions} dimensions, "\
              f"received array with {dimensions} dimensions."
        super().__init__(msg)


class DataStoreTypeError(Error):
    """The exception raised when anything other than a :class:`~numpy.ndarray`
    is passed to the data store.

    """
    def __init__(self, recieved_type: type):
        """The constructor for the :class:`DataStoreTypeError` class.

        """
        msg = f"Expecting {ndarray}, received {recieved_type}."
        super().__init__(msg)


class DataStoreGeometryError(Error):
    """The exception raised when the data store is queries with an geometry string
    it does not recognise.

    """
    def __init__(self, geometry_str: str):
        """The constructor for the :class:`DataStoreGeometryError` class.

        """
        msg = f"The  geometry '{geometry_str}' does not exist."
        super().__init__(msg)




from os.path import join

from io import TextIOWrapper

from typing import Dict
from typing import List
from typing import Type 

from ridt.base import RIDTOSError
from ridt.base import ComputationalSpace

from ridt.config import RIDTConfig
from ridt.config import Units
from ridt.config import summary

from ridt.container import Domain

from .resultcontainers import ResultContainer
from .datastoreanalyser import DataStoreAnalyser

FIRST = 0


class BatchResultsWriter:
    """Batch results writer class.

    Iterates through a :obj:`dict` of :class:`~.DataStoreAnalyser` instances
    and computes batch run analysis quantities.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.
    
    outdir: :obj:`str`
        The path to the output directory for the run.
    
    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    analysis : :obj:`dict` [:class:`~.RIDTConfig`, :class:`~.DataStoreAnalyser`]
        A dictionary of :class:`~.DataStoreAnalyser` instances for each
        :class:`~.RIDTConfig` object in :attr:`space`.
    
    thresholds : :obj:`list` [:obj:`float`]
        The threshold values corresponding to :attr:`quantity` defined in
        :attr:`setting`.
    
    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.

    """
    def __init__(self,
                 setting: RIDTConfig,
                 space: ComputationalSpace,
                 analysis: Dict[RIDTConfig, DataStoreAnalyser],
                 outdir: str,
                 quantity: str):
        """The :class`~.BatchResultsWriter` class initialiser.

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.

        space : :class:`~.ComputationalSpace`
            The :class:`~.ComputationalSpace` instance corresponding to the
            :attr:`settings` attribute.
        
        analysis : :obj:`dict` [:class:`~.RIDTConfig`, :class:`~.DataStoreAnalyser`]
            A dictionary of :class:`~.DataStoreAnalyser` instances for each
            :class:`~.RIDTConfig` object in :attr:`space`.

        outdir : :obj:`str`
            The path to the output directory for the run.

        quantity : :obj:`str`
            The string id for the quantity stored in the data  store.

       """
        self.setting = setting
        self.domain = Domain(setting)
        self.units = Units(setting)
        self.quantity = quantity
        self.thresholds = self.threshold_converter()
        self.space = space
        self.analysis = analysis
        self.outdir = outdir
        self.summary()
        self.write()

    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`setting`.

        """
        locations = self.setting.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def threshold_converter(self):
        """Converts the threshold into SI units.

        Converts the thresholds in :attr:`setting` corresponding to
        :attr:`quantity` into SI units.

        Returns
        -------
        :obj:`list` [:obj:`float`]
            The list of threshold values in SI units. 
        """
        tld = [t.value for t in getattr(self.setting.thresholds, self.quantity)]
        return getattr(self.units, f"{self.quantity}_converter")(tld)

    def get_valid(self, geometry: str, results: List[Type[ResultContainer]]):
        return [i for i in results if i.geometry == geometry and i.valid]

    def write(self):
        """Computes and writes batch results to disk.

        loops through all items in :attr:`analysis` and for each geometry type
        writes the following information to the output file:

        The maximum value.

        The fastest time to all threshold values.

        The fastest time to the defined percentage of domain to all threshold
        values.

        The largest percentage that exceeds all threshold values.
        The 

        Each item for each geometry is written indicating which settings object
        it corresponds to.

        Raises
        ------
        :class:`~.RIDTOSError`
            If unable to create the output file on disk.

        """
        try:
            fname = f"batch_{self.quantity}_extrema.txt"
            f = open(join(self.outdir, fname), 'w')
        except OSError as e:
            raise RIDTOSError(e)

        for geometry in self.geometries:
            over = list()
            for setting, analysis in self.analysis.items():
                items = self.get_valid(geometry, analysis.maximum)
                if items:
                    item = max(items)
                    over.append(item)
            if over:
                item = max(over)
                self.subtitle(f, item.extreme_title)
                f.write(f"run id: {self.space.index(item.setting)}\n")
                f.write(item.string)

        results = [
            ("exceedance", False),
            ("percent_exceedance", False),
            ("max_percent_exceedance", True)
        ]

        for r, reverse in results:
            for t in self.thresholds:
                for geometry in self.geometries:
                    over = list()
                    for setting, analysis in self.analysis.items():
                        valid = self.get_valid(geometry, getattr(analysis, r))
                        items = [i for i in valid if i.threshold == t]
                        if items:
                            item = min(items) if not reverse else max(items)
                            over.append(item)
                    if over:
                        item = min(over) if not reverse else max(over)
                        self.subtitle(f, item.extreme_title)
                        f.write(f"run id: {self.space.index(item.setting)}\n")
                        f.write(item.string)
        f.close()
    
    def summary(self):
        """Writes the run summary to the output folder.

        Raises
        ------
        :class:`~.RIDTOSError`  
            If unable to create the output file on disk.

        """
        try:
             f = open(join(self.outdir, "batch_run_summary.txt"), 'w')
        except OSError as e:
            raise RIDTOSError(e)

        f.write(self.space.cout_summary())
        f.write("\n")
        f.write(summary(self.setting))
        f.close()

    def title(self, file: TextIOWrapper, title: str):
        """Writes a formatted title string to a file object.

        Parameters
        ----------
        file : :obj:`TextIOWrapper` 
            The file object to write to.

        title : :obj:`str`
            The title string.

        Returns
        -------
        None

        """
        file.write("\n")
        file.write("".join("=" for i in range(len(title))) + "\n")
        file.write(title + "\n")
        file.write("".join("=" for i in range(len(title))) + "\n")

    def subtitle(self, file: TextIOWrapper, subtitle: str):
        """Writes a formatted subtitle string to a file object.

        Parameters
        ----------
        file : :obj:`TextIOWrapper` 
            The file object to write to.

        subtitle : :obj:`str`
            The subtitle string.
        
        Returns
        -------
        None
        """
        file.write("\n")
        file.write("".join("-" for i in range(len(subtitle))) + "\n")
        file.write(subtitle + "\n")
        file.write("".join("-" for i in range(len(subtitle))) + "\n")



import unittest
import json

import os

from ridt.config import ConfigFileParser
from ridt.config.ridtconfig import ConsistencyError
from ridt.config.configfileparser import ConfigFileParserValidationError


class ST06(unittest.TestCase):

    """System Test 06. Test that the system can parse
     all physical unit specifications from the JSON file."""

    def setUp(self) -> None:

        """setUp method that instantiates the
        :class:`~.RIDTConfig` class."""


        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.config_path = os.path.join(this_dir, "st06/config.json")
        with open(self.config_path) as f:
            self.default = json.load(f)

        this_dir = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(this_dir, "../../default/config.json")
        with ConfigFileParser(path) as cfp:
            self.c = cfp

    def tearDown(self) -> None:

        with open(self.config_path, "w") as w:
            json.dump(self.default, w)

    def test_verify(self):

        """Verifies that the :class:`~.RIDTConfig` class
        contains the correct the attributes."""

        self.assertEqual(
            hasattr(self.c, "time_units"), True)
        self.assertEqual(
            hasattr(self.c, "concentration_units"), True)
        self.assertEqual(
            hasattr(self.c, "exposure_units"), True)

    def test_concentration_units(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["concentration_units"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_spatial_units(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["spatial_units"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_exposure_units(self):

        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["exposure_units"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_mass_units(self):
        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["mass_units"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)

    def test_time_units(self):
        with self.assertRaises(ConfigFileParserValidationError):
            with open(self.config_path) as f:
                config = json.load(f)
            config["time_units"] = "test"
            with open(self.config_path, "w") as f:
                json.dump(config, f)
            ConfigFileParser(self.config_path)


if __name__ == "__main__":
    unittest.main()



from typing import Tuple

from ridt.container import Domain

from ridt.config import RIDTConfig
from ridt.config import Units


class ResultContainer:
    """A base class for all results containers.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.
    
    geometry : :obj:`str`
        The geometry domain this result corresponds to.

    id : :obj:`str`
        The id of the domain this result corresponds to.
    
    domain : :class:`~.Domain`
        The instance of :class:`~.Domain` corresponding to :attr:`setting`.

    quantity: :obj:`str`
        The string id for the quantity stored in the data  store.

    units : :class:`~.Units`
        The instance of :class:`~.Units` corresponding to :attr:`setting`.

    """


    def __init__(self, setting: RIDTConfig, geometry: str, id: str, quantity: str):
        self.setting = setting
        self.geometry = geometry
        self.id = id
        self.quantity = quantity
        self.units = Units(setting)
        self.domain = Domain(setting)
    
    def same_geometry(self, other):
        """Checks if item has same :attr:`geometry`.

        Parameters
        ----------
        other
            The comparative item.

        Raises
        ------
        :obj:`ValueError`
            if the :attr:`geometry` attributes are different.
        """
        if self.geometry != other.geometry:
            raise ValueError("You are comparing two different geometries.")
    
    @property
    def unit(self):
        """:obj:`str` The units corresponding :attr:`quantity`.

        """
        return getattr(self.units, f"{self.quantity}")
    

class Maximum(ResultContainer):
    """Result container for the max value in a domain.

    Attributes
    ----------
    index : Tuple[:obj:`int`]
        The index in the array where the maximum occurred.

    value : :obj:`float`    
        The value of the maximum.

    """

    def __init__(self,
                 setting: RIDTConfig,
                 geometry: str,
                 id: str,
                 quantity: str,
                 index: Tuple[int],
                 value: float):
        """The Maximum class initialiser

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        geometry : :obj:`str`
            The geometry domain this result corresponds to.

        id : :obj:`str`
            The id of the domain this result corresponds to.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        index : Tuple[:obj:`int`]
            The index in the array where the maximum occurred.
    
        value : :obj:`float`    
            The value of the maximum.

        """
        super().__init__(setting, geometry, id, quantity)
        self.index = index
        self.value = value
    
    def __lt__(self, other):
        if not isinstance(other, Maximum):
            raise TypeError(
                f"< not supported between instances of {Maximum} and {type(other)}")
        self.same_geometry(other)
        return True if self.value < other.value else False
    
    @property
    def string(self):
        """:obj:`str` : The string representation of the result.

        """
        u = getattr(self.units, f"{self.quantity}")
        factor = getattr(self.units, f"{self.quantity}_factor")
        rv = str()
        if self.index:
            t, x, y, z = self.domain.values(self.id, self.index)
            rv += f"id: {self.id}\n"
            rv += f"time: {t:.2f}{self.units.time}\n"
            rv += f"x: {x:.2f}{self.units.space}\n"
            rv += f"y: {y:.2f}{self.units.space}\n"
            rv += f"z: {z:.2f}{self.units.space}\n"
            rv += f"value: {self.value / factor:.2e}{u}\n\n"
        else:
            rv += "None\n\n"
        return rv
    
    @property
    def header(self):
        """:obj:`list` [:obj:`str`] : The list of headers for a csv output.

        """
        return [
            "id",
            f"time ({self.units.time})",
            f"x ({self.units.space})",
            f"y ({self.units.space})",
            f"z ({self.units.space})",
            f"value ({self.unit})"
        ]
    
    @property
    def row(self):
        """:obj:`list` [:obj:`float`] : The list of values for a csv output.

        """
        factor = getattr(self.units, f"{self.quantity}_factor")
        if self.index:
            t, x, y, z = self.domain.values(self.id, self.index)
            return [self.id, t, x, y, z, self.value / factor]
        else:
            return [self.id, "None", "None", "None", "None"]
    
    @property
    def fname(self):
        """:obj:`str` : the file name for the this result output.

        """
        return f"{self.geometry}_maximums.csv"

    @property
    def title(self):
        """:obj:`str` : the title string for this result.

        """
        return "Maxima"

    @property
    def extreme_title(self):
        """:obj:`str` : the extremum version of the title for this result.

        """
        return f"Maximum value for {self.geometry}:"
    
    @property
    def valid(self):
        """:obj:`bool` : Returns true if contains valid index, else false.

        """
        if self.index and self.value != "nan":
            return True
        else:
            return False
    

class Exceedance(ResultContainer):
    """Result container for the threshold exceedance in a domain.

    Attributes
    ----------
    index : Tuple[:obj:`int`]
        The index in the array where the exceedance occurred.

    threshold : :obj:`float`
        The threshold that was exceeded.

    """

    def __init__(self,
                 setting: RIDTConfig,
                 geometry: str,
                 id: str,
                 quantity: str,
                 index: Tuple[int],
                 threshold: float):
        """The Exceedance class initialiser

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        geometry : :obj:`str`
            The geometry domain this result corresponds to.

        id : :obj:`str`
            The id of the domain this result corresponds to.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        index : Tuple[:obj:`int`]
            The index in the array where the exceedance occurred.
    
        threshold : :obj:`float`
            The threshold that was exceeded.

        """
        super().__init__(setting, geometry, id, quantity)
        self.threshold = threshold
        self.index = index
    
    def __lt__(self, other):
        if not isinstance(other, Exceedance):
            raise TypeError(
        f"< not supported between instances of {Exceedance} and {type(other)}")
        self.same_geometry(other)
        return True if self.index[0] < other.index[0] else False
    
    @property
    def string(self):
        """:obj:`str` : The string representation of the result.

        """
        rv = str()
        if self.index:
            t, x, y, z = self.domain.values(self.id, self.index)
            rv += f"id: {self.id}\n"
            rv += f"time: {t:.2f}{self.units.time}\n"
            rv += f"x: {x:.2f}{self.units.space}\n"
            rv += f"y: {y:.2f}{self.units.space}\n"
            rv += f"z: {z:.2f}{self.units.space}\n"
        else:
            rv += "None\n\n"
        return rv
    
    @property
    def header(self):
        """:obj:`list` [:obj:`str`] : The list of headers for a csv output.

        """
        return [
            "id",
            f"time ({self.units.time})",
            f"x ({self.units.space})",
            f"y ({self.units.space})",
            f"z ({self.units.space})",
        ]

    @property
    def row(self):
        """:obj:`list` [:obj:`float`] : The list of values for a csv output.

        """
        if self.index:
            t, x, y, z = self.domain.values(self.id, self.index)
            return [self.id, t, x, y, z]
        else:
            return [self.id, "None", "None", "None"]
    
    @property
    def fname(self):
        """:obj:`str` : the file name for the this result output.

        """
        return f"{self.geometry}_exceeds_{self.threshold}{self.unit}.csv"

    @property
    def title(self):
        """:obj:`str` : the title string for this result.

        """
        return "Threshold Exceedance"
    
    @property
    def extreme_title(self):
        """:obj:`str` : the extremum version of the title for this result.

        """
        return f"Minimum time to {self.threshold}{self.unit} for {self.geometry}:"
    
    @property
    def valid(self):
        """:obj:`bool` : Returns true if contains valid index, else false.

        """
        return True if self.index else False


class PercentExceedance(ResultContainer):
    """Result container for the percent threshold exceedance in a domain.

    Attributes
    ----------
    index : :obj:`int`
        The time index in the array where the exceedance occurred.

    threshold : :obj:`float`
        The threshold that was exceeded.
    
    percent : :obj:`float`
        The percentage by which the threshold was exceeded

    """

    def __init__(self,
                 setting: RIDTConfig,
                 geometry: str,
                 id: str,
                 quantity: str,
                 index: int,
                 threshold: float,
                 percent: float):
        """The PercentExceedance class initialiser

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        geometry : :obj:`str`
            The geometry domain this result corresponds to.

        id : :obj:`str`
            The id of the domain this result corresponds to.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        index : :obj:`int`
            The time index in the array where the exceedance occurred.
    
        threshold : :obj:`float`
            The threshold that was exceeded.

        percent : :obj:`float`
            The percentage by which the threshold was exceeded

        """
        super().__init__(setting, geometry, id, quantity)
        self.threshold = threshold
        self.index = index
        self.percent = percent
    
    def __lt__(self, other):
        if not isinstance(other, PercentExceedance):
            raise TypeError(
                f"< not supported between instances of {PercentExceedance} and"\
                f" {type(other)}")
        self.same_geometry
        return True if self.index < other.index else False
    
    @property
    def string(self):
        """:obj:`str` : The string representation of the result.

        """
        rv = str()
        if self.index:
            t = self.domain.time[self.index]
            rv += f"id: {self.id}\n"
            rv += f"time: {t:.2f}{self.units.time}\n\n"
        else:
            rv += "None\n\n"
        return rv
    
    @property
    def header(self):
        """:obj:`list` [:obj:`str`] : The list of headers for a csv output.

        """
        return [
            "id",
            f"time ({self.units.time})",
        ]

    @property
    def row(self):
        """:obj:`list` [:obj:`float`] : The list of values for a csv output.

        """
        if self.index:
            t = self.domain.time[self.index]
            return [self.id, t]
        else:
            return [self.id, "None"]
    
    @property
    def fname(self):
        """:obj:`str` : the file name for the this result output.

        """
        return f"{self.percent}%_of_{self.geometry}_exceeds_{self.threshold}{self.unit}.csv"
    
    @property
    def title(self):
        """:obj:`str` : the title string for this result.

        """
        return f"{self.percent}% Threshold Exceedance"

    @property
    def extreme_title(self):
        """:obj:`str` : the extremum version of the title for this result.

        """
        return f"Minimum time to {self.threshold}{self.unit} for "\
               f"{self.percent}% of domain for {self.geometry}:"

    @property
    def valid(self):
        """:obj:`bool` : Returns true if contains valid index, else false.

        """
        return True if self.index else False


class MaxPercentExceedance(ResultContainer):
    """Result container for the max percent threshold exceedance in a domain.

    Attributes
    ----------
    index : :obj:`int`
        The time index in the array where the exceedance occurred.

    threshold : :obj:`float`
        The threshold that was exceeded.
    
    value : :obj:`float`
        The percentage by which the threshold was exceeded


    """
    def __init__(self,
                 setting: RIDTConfig,
                 geometry: str,
                 id: str,
                 quantity:str,
                 value: float,
                 index: int,
                 threshold: float):
        """The PercentExceedance class initialiser

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        geometry : :obj:`str`
            The geometry domain this result corresponds to.

        id : :obj:`str`
            The id of the domain this result corresponds to.
        
        quantity: :obj:`str`
            The string id for the quantity stored in the data  store.

        index : :obj:`int`
            The time index in the array where the exceedance occurred.
    
        threshold : :obj:`float`
            The threshold that was exceeded.

        value: :obj:`float`
            The percentage by which the threshold was exceeded
        
        """
        super().__init__(setting, geometry, id, quantity)
        self.threshold = threshold
        self.index = index
        self.value = value

    def __lt__(self, other):
        if not isinstance(other, MaxPercentExceedance):
            raise TypeError(
        f"< not supported between instances of {MaxPercentExceedance} and {type(other)}")
        self.same_geometry(other)
        return True if self.value < other.value else False
    
    @property
    def string(self):
        """:obj:`str` : The string representation of the result.

        """
        rv = str()
        if self.index:
            t = self.domain.time[self.index]
            rv += f"id: {self.id}\n"
            rv += f"time: {t:.2f}{self.units.time}\n"
            rv += f"value: {self.value:.2f}%\n\n"
        else:
            rv += "None\n\n"
        return rv
    
    @property
    def header(self):
        """:obj:`list` [:obj:`str`] : The list of headers for a csv output.

        """
        return [
            "id",
            f"time ({self.units.time})",
            f"value (%)"
        ]

    @property
    def row(self):
        """:obj:`list` [:obj:`float`] : The list of values for a csv output.

        """
        if self.index:
            t = self.domain.time[self.index]
            return [self.id, t, self.value]
        else:
            return [self.id, "None", "None"]

    @property
    def fname(self):
        """:obj:`str` : the file name for the this result output.

        """
        return f"{self.geometry}_max%_exceeds_{self.threshold}{self.unit}.csv"

    @property
    def title(self):
        """:obj:`str` : the title string for this result.

        """
        return "Maximum % Threshold Exceedance"
    
    @property
    def extreme_title(self):
        """:obj:`str` : the extremum version of the title for this result.

        """
        return f"Maximum percentage exceeding {self.threshold}{self.unit} "\
               f"for {self.geometry}:"

    @property
    def valid(self):
        """:obj:`bool` : Returns true if contains valid index, else false.

        """
        return True if self.index else False



from ridt.base import ComputationalSpace
from ridt.equation import WellMixed

from ridt.config import RIDTConfig

from ridt.data import BatchDataStore
from ridt.data import BatchDataStoreWriter
from ridt.data import BatchDataStorePlotter

from ridt.analysis import Exposure

from ridt.container import Domain


BF = '{l_bar}{bar:30}{r_bar}{bar:-10b}'


class WellMixedRun:
    """The class which orchestrates an Well Mixed model run.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.

    data_store : :class:`~.BatchDataStore`
        The batch run data store for concentration values.

    outdir: :obj:`str`
        The path to the output directory for the run.
    
    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.
    
    exposure_store : :obj:`Union`[:class:`~.BatchDataStore`, None]
        The batch run data store for exposure values.
    
    """
    def __init__(self, settings: RIDTConfig, output_dir: str):
        """The constructor for the :class:`EddyDiffusion` class.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        outdir : :obj:`str`
            The path to the output directory for the run.

        """
        self.settings = settings
        self.outdir = output_dir
        self.data_store = BatchDataStore()
        self.exposure_store = None
        self.space = self.prepare()
        self.evaluate()
        self.compute_exposure()
        self.write()
        self.plot()
        print("\n")

    def prepare(self) -> ComputationalSpace:
        """Instantiates :class:`~.ComputationalSpace` instance.

        Generates a :class:`~.ComputationalSpace` instance from
        :attr:`settings`. 

        Returns
        -------
        :class:`~.ComputationalSpace`
            The Computational Space created from :attr:`settings`. 
        """
        print("Preparing Well Mixed run... ")
        restrict = {"models": "well_mixed"}
        return ComputationalSpace(self.settings, restrict)
    
    def evaluate(self) -> None:
        """Loops over all elements in :attr:`space` and evaluates the model.

        Returns
        -------
        None

        """
        print("Evaluating model over domain... ")
        for setting in self.space.space:
            count = f"{self.space.linear_index(setting) + 1}/{len(self.space)}"
            print(f"Evaluating computational space element {count}")
            self.run(setting)

    def run(self, setting: RIDTConfig) -> None:
        """Evaluates the model for a set of parameters.
    
        Writes output to :attr:`data_store`.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.
        
        Returns
        -------
        None

        """
        self.data_store.add_run(setting)
        domain = Domain(setting)
        solver = WellMixed(setting)
        output = solver(domain.time)
        self.data_store[setting].add("points", "well_mixed", output)
    
    def compute_exposure(self):
        """Computes the exposure from the concentration data.

        Saves the result to a new data store :attr:`exposure_store`

        Returns
        -------
        None

        """
        if self.settings.compute_exposure:
            print("Computing exposure...")
            self.exposure_store = Exposure(self.settings, self.data_store)

    def args(self, data_store: BatchDataStore, quantity: str) -> tuple:
        """A helper function that returns a tuple of some values.

        Parameters
        ----------

        data_store : :class:`~.BatchDataStore`
            A batch data store.

        quantity : :obj:`str`
            The name of the quantity in the data store.

        Returns
        -------
        :obj:`tuple`
            The various quantities as a tuple.

        """
        return self.settings, data_store, self.space, self.outdir, quantity

    def write(self) -> None:
        """Writes all data stores to disk.

        Returns
        -------
        None

        """
        print("\nWriting data to disk... ")
        BatchDataStoreWriter(*self.args(self.data_store, "concentration"))
        if self.settings.compute_exposure:
            BatchDataStoreWriter(*self.args(self.exposure_store, "exposure"))

    def plot(self) -> None:
        """Plots all requested data and writes it to disk.

        Returns
        -------
        None

        """
        print("\nProducing plots... ")
        BatchDataStorePlotter(*self.args(self.data_store, "concentration"))
        if self.settings.compute_exposure:
            BatchDataStorePlotter(*self.args(self.exposure_store, "exposure"))



import unittest

import os

import numpy as np

from ridt.config import ConfigFileParser

from ridt.equation import EddyDiffusion


class ST16(unittest.TestCase):

    """System Test 16. Test that the system can
       evaluate the TKEB diffusion coefficient
       and use it in the Eddy Diffusion model."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        with ConfigFileParser(os.path.join(this_dir, "st16/explicit.json")) as cfp:
            self.e = cfp
        with ConfigFileParser(os.path.join(this_dir, "st16/tkeb.json")) as cfp:
            self.t = cfp

        self.explicit = EddyDiffusion(self.e)
        self.tkeb = EddyDiffusion(self.t)

    def test_verify(self):
        self.assertEqual(
            self.explicit.diffusion_coefficient(), 0.001
        )
        self.assertEqual(
            self.tkeb.diffusion_coefficient(), 0.7655
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from ridt.data.datastore import DataStore

from ridt.data.datastore import DataStoreDimensionalityError
from ridt.data.datastore import DataStoreTypeError
from ridt.data.datastore import DataStoreIDError


class TestDataStore(unittest.TestCase):

    """Unit tests for the :class:`~.DataStore` class."""

    def setUp(self) -> None:

        """setUp method which instantiates the :class:`~.DataSore` class,
        and creates all the data for the tests."""

        self.ds = DataStore()
        self.space = np.linspace(0, 10, 10)

        self.point_data = np.array(self.space)

        self.line_data = np.array(
            [self.space, self.space])

        self.plane_data = np.array(
            [[self.space, self.space], [self.space, self.space]])

        self.domain_data = np.array(
            [[[self.space, self.space], [self.space, self.space]],
             [[self.space, self.space], [self.space, self.space]]]
        )



from typing import Iterable

from .ridtconfig import RIDTConfig
from .units import Units
from ridt.equation.eddy_diffusion import EddyDiffusion


def summary(settings: RIDTConfig) -> str:
    """A function that generates a formatted string containing run info.

    This takes a settings object and generates a string with various information
    about the run. It also computes various sanity check quantities.

    Parameters
    ----------
    settings : :class:`~.RIDTConfig`
        The settings object for the run in question. 

    Returns
    -------
    :obj:`str` 
        The formatted string, ready to be written to a file.
    """
    units = Units(settings)
    rv = str()
    rv += f"RIDT v1.0\n"

    rv += f"Models:\n"
    rv += f"\tEddy Diffusion: {settings.eddy_diffusion}\n"
    rv += f"\tWell Mixed: {settings.well_mixed}\n"
    rv += f"\n"

    rv += f"total time: {settings.total_time} {settings.time_units}\n"
    rv += f"\n"
    rv += f"spatial_dimensions:\n"
    rv += f"\tx: {settings.dimensions.x} {units.space}\n"
    rv += f"\ty: {settings.dimensions.y} {units.space}\n"
    rv += f"\tz: {settings.dimensions.z} {units.space}\n"
    rv += f"\n"
    rv += f"fresh air change rate: {settings.fresh_air_flow_rate} {settings.fresh_air_flow_rate_units}\n"
    rv += f"\n"

    if settings.eddy_diffusion:
        coeff = settings.models.eddy_diffusion.coefficient
        rv += f"\n"
        if coeff.calculation == "TKEB":
            rv += f"diffusion coefficient: TKEB\n"
            rv += f"\tvents: {coeff.tkeb.number_of_supply_vents}\n"
            rv += f"\ttotal air change rate: {coeff.tkeb.total_air_flow_rate} m3.s-1\n"
            M = EddyDiffusion(settings)
            rv += f"diffusion coefficient value: {M.diffusion_coefficient()} m2.s-1\n"
        else:
            rv += f"diffusion coefficient value: {coeff.value} m2.s-1\n"
        if settings.models.eddy_diffusion.images.mode == "manual":
            rv += f"number of image sources computed (either side of zero): \
                {settings.models.eddy_diffusion.images.quantity}"
        else:
            rv += f"number of image sources computed: auto."
        rv += f"\n"
    if settings.well_mixed:
        dim = settings.dimensions
        if isinstance(dim.x, list) or isinstance(dim.y, list) or isinstance(dim.z, list):
            rv += f"well mixed volume: <various> m3\n"
        else:
            rv += f"well mixed volume: {settings.dimensions.x * settings.dimensions.y *  settings.dimensions.z} m3\n"
        rv += f"\n"

    rv += "sources\n"
    rv += "-------\n"

    rv += "instantaneous:\n"
    for name, item in settings.modes.instantaneous.sources.items():
        rv += f"\t{name}\n"
        rv += f"\t\tx: {item.x} {units.space}\n"
        rv += f"\t\ty: {item.y} {units.space}\n"
        rv += f"\t\tz: {item.z} {units.space}\n"
        rv += f"\t\tmass: {item.mass} {units.mass}\n"
        rv += f"\t\ttime: {item.time} {units.time}\n"
        dim = settings.dimensions
        if isinstance(item.mass, list) or\
           isinstance(dim.x, list) or\
           isinstance(dim.y, list) or\
           isinstance(dim.z, list):
            rv += f"\t\tsteady state well mix concentration: <various>\n"
        else:
            vol = settings.dimensions.x * settings.dimensions.y * settings.dimensions.z
            rv += f"\t\tsteady state well mix concentration: {(item.mass / vol):.2e} {units.concentration_si}\n"
        if isinstance(item.mass, list) or\
           isinstance(settings.fresh_air_flow_rate, list):
            rv += f"\t\tupper exposure limit: <various>\n"
        else:
            if settings.fresh_air_flow_rate:
                rv += f"\t\tupper exposure limit: \
                    {(item.mass / settings.fresh_air_flow_rate):.2e} {units.exposure_si}\n"
            else:
                rv += f"\t\tupper exposure limit: infinite.\n"


        rv += f"\n"
    rv += "infinite duration:\n"
    for name, item in settings.modes.infinite_duration.sources.items():
        rv += f"\t{name}\n"
        rv += f"\t\tx: {item.x} {units.space}\n"
        rv += f"\t\ty: {item.y} {units.space}\n"
        rv += f"\t\tz: {item.z} {units.space}\n"
        rv += f"\t\trate: {item.rate} kg.m-3\n"
        rv += f"\t\ttime: {item.time} {units.time}\n"
        if isinstance(item.rate, list) or\
           isinstance(settings.fresh_air_flow_rate, list):
            rv += f"\t\tsteady state well mix concentration: <various>\n"
        else:
            if settings.fresh_air_flow_rate:
                rv += f"\t\tsteady state well mix concentration: \
                    {(item.rate / settings.fresh_air_flow_rate):.2e} {units.concentration_si}\n"
            else:
                rv += f"\t\tsteady state well mix concentration: infinite\n"

        rv += f"\n"
    rv += "fixed duration:\n"
    for name, item in settings.modes.fixed_duration.sources.items():
        rv += f"\t{name}\n"
        rv += f"\t\tx: {item.x} {units.space}\n"
        rv += f"\t\ty: {item.y} {units.space}\n"
        rv += f"\t\tz: {item.z} {units.space}\n"
        rv += f"\t\trate: {item.rate} kg.m-3\n"
        rv += f"\t\tstart_time: {item.end_time} {units.time}\n"
        rv += f"\t\tend_time: {item.end_time} {units.time}\n"
        dim = settings.dimensions
        if isinstance(item.rate, list) or\
           isinstance(dim.x, list) or\
           isinstance(dim.y, list) or\
           isinstance(dim.z, list) or\
           isinstance(item.start_time, list) or\
           isinstance(item.end_time, list):
            rv += f"\t\tsteady state well mix concentration: <various>\n"
        else:
            vol = settings.dimensions.x * settings.dimensions.y * settings.dimensions.z
            rv += f"\t\tsteady state well mix concentration: \
                {(item.rate * (item.end_time - item.end_time) / vol):.2e} {units.concentration_si}\n"
        if isinstance(item.rate, list) or\
           isinstance(item.start_time, list) or\
           isinstance(item.end_time, list) or\
           isinstance(settings.fresh_air_flow_rate, list):
            rv += f"\t\tupper exposure limit: <various>\n"
        else:
            if settings.fresh_air_flow_rate:
                rv += f"\t\tupper exposure limit: \
                    {(item.rate * (item.end_time - item.start_time) / settings.fresh_air_flow_rate):.2e} {units.exposure_si}\n"
            else:
                rv += f"\t\tupper exposure limit: infinite\n"

        rv += f"\n"
    
    rv += "\n"
    rv += "monitor locations\n"
    rv += "-----------------\n"

    rv += "points:\n"
    for name, item in settings.models.eddy_diffusion.monitor_locations.points.items():
        rv += f"\t{name}\n"
        rv += f"\t\tx: {item.x} {units.space}\n"
        rv += f"\t\ty: {item.y} {units.space}\n"
        rv += f"\t\tz: {item.z} {units.space}\n"
        rv += f"\n"
    
    rv += "lines:\n"
    for name, item in settings.models.eddy_diffusion.monitor_locations.lines.items():
        rv += f"\t{name}\n"
        rv += f"\t\tpoint - x: {item.point.x} {units.space}\n"
        rv += f"\t\tpoint - y: {item.point.y} {units.space}\n"
        rv += f"\t\tpoint - z: {item.point.z} {units.space}\n"
        rv += f"\t\tparallel axis: {item.parallel_axis}\n"
        rv += f"\n"
    
    rv += "planes:\n"
    for name, item in settings.models.eddy_diffusion.monitor_locations.planes.items():
        rv += f"\t{name}\n"
        rv += f"\t\tplane: {item.axis}\n"
        rv += f"\t\tposition: {item.distance} {units.space}\n"
        rv += f"\n"

    return rv


import csv
import sys

from os.path import basename
from os.path import dirname

from ridt.base import Error

from ridt.config import ConfigFileParser
from ridt.config import ConfigFileWriter


class CSVToConfigFile:

    """Class which parsed a csv file and adds the entries to a config file.

    The CSV file must be of the following format:

    INS,float,float,float,float,float,
    INF,float,float,float,float,float,
    FIX,float,float,float,float,float,float
    POI,float,float,float,,,
    LIN,float,float,float,x,,
    PLA,xy,float,,,,

    Where many of each type of entry can exist.

    INS - instantaneous source.
    INF - infinite duration source.
    FIX - fixed duration
    POI - monitor point
    LIN - monitor line
    PLA - monitor plane.

    Attributes
    ----------
    csv_file_path : :obj:`str`
        The path to the csv file to be parsed.

    config_file_path : :obj:`str`
        The path to the config file to be parsed.
    
    output_file_path : :obj:`str`
        The path to save the new config file.
    
    parsed_items : :obj:`dict` [:obj:`str`, :obj:`list`]
        The items parsed from the CSV file.
    
    config : :class:`~.RIDTConfig`
        The instantiated settings object.
    
    new : :obj:`dict`
        The source dictionary for the old and new settings object.

    """

    def __init__(self):
        pass

    def __call__(self,
                 config_file_path: str,
                 csv_file_path: str,
                 output_file_path: str):
        """The CSVToConfigFile initialiser

        Parameters
        ----------
        csv_file_path : :obj:`str`
            The path to the csv file to be parsed.

        config_file_path : :obj:`str`
            The path to the config file to be parsed.
        
        output_file_path : :obj:`str`
            The path to save the new config file.
        
        """
        self.csv_file_path = csv_file_path
        self.config_file_path = config_file_path
        self.output_file_path = output_file_path
        self.parsed_items = {
            "INS": [],
            "INF": [],
            "FIX": [],
            "POI": [],
            "LIN": [],
            "PLA": []
        }
        self.config = ConfigFileParser(config_file_path)
        self.new = self.config.__source__
        self.parse_csv()
        self.write_new_config_file()

    def parse_csv(self):
        """Open and parse the entries in the CSV file.

        It adds the items to :attr:`parsed_items`

        Raises
        ------
        :class:`~.CSVToConfigFileOSError`
            If the csv file cannot be opened.

        """
        try:
            with open(self.csv_file_path, 'r') as f:
                c = csv.reader(f, delimiter=",")
                for row in c:
                    if row[0]:
                        if row[0] in self.parsed_items:
                            self.parsed_items[row[0]].append(
                                [item for item in row[1:] if item != ""])
                        else:
                            print(f"Invalid spec: {row[0]}")
        except OSError as e:
            raise CSVToConfigFileOSError(self.csv_file_path, e)
        try:
            self.add_INS()
            self.add_INF()
            self.add_FIX()
            self.add_POI()
            self.add_LIN()
            self.add_PLA()
        except CSVToConfigFileValueError as e:
            sys.exit(e)
        except CSVToConfigFileIndexError as e:
            sys.exit(e)

    def add_INS(self):
        """Add the parsed INS items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.

        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["INS"]):
            try:
                self.new["modes"]\
                        ["instantaneous"]\
                        ["sources"]\
                        [f"source_{idx}"] = {
                    "x": float(item[0]),
                    "y": float(item[1]),
                    "z": float(item[2]),
                    "mass": float(item[3]),
                    "time": float(item[4])
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("INS", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("INS", self.ord(idx), 5)

    def add_INF(self):
        """Add the parsed INF items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.

        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["INF"]):
            try:
                self.new["modes"]\
                        ["infinite_duration"]\
                        ["sources"]\
                        [f"source_{idx}"] = {
                    "x": float(item[0]),
                    "y": float(item[1]),
                    "z": float(item[2]),
                    "rate": float(item[3]),
                    "time": float(item[4])
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("INF", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("INF", self.ord(idx), 5)

    def add_FIX(self):
        """Add the parsed FIX items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.

        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["FIX"]):
            try:
                self.new["modes"]\
                        ["fixed_duration"]\
                        ["sources"]\
                        [f"source_{idx}"] = {
                    "x": float(item[0]),
                    "y": float(item[1]),
                    "z": float(item[2]),
                    "rate": float(item[3]),
                    "start_time": float(item[4]),
                    "end_time": float(item[5])
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("FIX", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("FIX", self.ord(idx), 6)
    
    def add_POI(self):
        """Add the parsed POI items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.

        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["POI"]):
            try:
                self.new["models"]\
                        ["eddy_diffusion"]\
                        ["monitor_locations"]\
                        ["points"]\
                        [f"point_{idx}"] = {
                    "x": float(item[0]),
                    "y": float(item[1]),
                    "z": float(item[2]),
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("POI", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("POI", self.ord(idx), 3)
    
    def add_LIN(self):
        """Add the parsed LIN items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.
            
        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["LIN"]):
            try:
                self.new["models"]\
                        ["eddy_diffusion"]\
                        ["monitor_locations"]\
                        ["lines"]\
                        [f"line_{idx}"] = {
                    "point": {
                        "x": float(item[0]),
                        "y": float(item[1]),
                        "z": float(item[2]),
                    },
                    "axis": item[3]
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("LIN", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("LIN", self.ord(idx), 6)
    
    def add_PLA(self):
        """Add the parsed PLA items to the source dictionary :attr:`new`

        Returns
        -------
        None

        Raises
        ------
        :class:`~.CSVToConfigFileValueError`
            If the parsed data is of the incorrect type.

        :class:`~.CSVToConfigFileIndexError`    
            If an incorrect number of items were in a given row parsed from the
            CSV file.

        """
        for idx, item in enumerate(self.parsed_items["PLA"]):
            try:
                self.new["models"]\
                        ["eddy_diffusion"]\
                        ["monitor_locations"]\
                        ["planes"]\
                        [f"plane_{idx}"] = {
                    "axis": str(item[0]),
                    "distance": float(item[1])
                }
            except ValueError as e:
                raise CSVToConfigFileValueError("PLA", self.ord(idx))
            except IndexError as e:
                raise CSVToConfigFileIndexError("PLA", self.ord(idx), 2)
    
    def write_new_config_file(self):
        if not self.output_file_path:
            path = self.append_id(self.config_file_path)
        else:
            path = self.output_file_path
        
        ConfigFileWriter(dirname(path), basename(path), self.new)

    def ord(self, n: int):
        """Returns the correct order suffix for a given integer.

        Parameters
        ----------
        n : :obj:`int`
            The integer in question.

        Returns
        -------
        :obj:`str`
            The order suffix.
            
        """
        return str(n)+("th" if 4<=n%100<=20 else {1:"st", 2:"nd", 3:"rd"}\
            .get(n%10, "th"))

    def append_id(self, filename: str):
        """Add a 'new' suffix to the file name before the file extension.

        Parameters
        ----------
        filename : :obj:`str`
            The file name to be extended.

        Returns
        -------
        :obj:`str`
            The new file name.

        """
        return "{0}_{2}.{1}".format(*filename.rsplit('.', 1) + ["new"]) 

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class CSVToConfigFileOSError(Error):
    """The exception raised when a :class:`~.CSVToConfigFile` instance
    catches an error when parsing a csv file path.

    """
    def __init__(self, path: str, error: OSError):
        """The constructor for the :class:`CSVToConfigFileOSError` class.

        """
        msg = f"While loading {path}, I encountered the following error:"\
              f"\n {error}"
        super().__init__(msg)


class CSVToConfigFileValueError(Error):
    """The exception raised when a :class:`~.CSVToConfigFile` instance
    catches a ValueError error when parsing a csv file.

    """
    def __init__(self, geometry: str, index: int):
        """The constructor for the :class:`CSVToConfigFileValueError` class.

        """
        msg = f"While loading the {index} {geometry} item, one of the parsed "\
              f"items was of an incorrect type."
        super().__init__(msg)


class CSVToConfigFileIndexError(Error):
    """The exception raised when a :class:`~.CSVToConfigFile` instance
    catches an IndexError error when parsing a csv file.

    """
    def __init__(self, geometry: str, index: str, expected: int):
        """The constructor for the :class:`CSVToConfigFileIndexError` class.

        """
        msg = f"While loading the {index} {geometry} item, not enough "\
              f"values were found. Expecting {expected}"
        super().__init__(msg)



from numpy import squeeze 

from ridt.base import ComputationalSpace

from ridt.config import RIDTConfig

from ridt.equation import EddyDiffusion

from ridt.data import BatchDataStore
from ridt.data import BatchDataStoreWriter
from ridt.data import BatchDataStorePlotter

from ridt.container import Domain

from ridt.analysis import BatchDataStoreAnalyser
from ridt.analysis import Exposure


class EddyDiffusionRun:
    """The class which orchestrates an Eddy Diffusion model run.

    Attributes
    ----------
    setting : :class:`~.RIDTConfig`
        The settings for the run in question.

    data_store : :class:`~.BatchDataStore`
        The batch run data store for concentration values.

    outdir: :obj:`str`
        The path to the output directory for the run.
    
    space : :class:`~.ComputationalSpace`
        The :class:`~.ComputationalSpace` instance corresponding to the
        :attr:`settings` attribute.
    
    exposure_store : :obj:`Union`[:class:`~.BatchDataStore`, None]
        The batch run data store for exposure values.
    
    """
    def __init__(self, settings: RIDTConfig, outdir: str):
        """The constructor for the :class:`EddyDiffusionRun` class.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        outdir : :obj:`str`
            The path to the output directory for the run.

        """
        self.settings = settings
        self.outdir = outdir
        self.data_store = BatchDataStore()
        self.exposure_store = None
        self.space = self.prepare()
        self.evaluate()
        self.compute_exposure()
        self.write()
        self.plot()
        self.analyse()

    @property
    def geometries(self):
        """:obj:`list` [:obj:`str`] : the list of geometries selected for
        evaluation in :attr:`settings`.

        """
        locations = self.settings.models.eddy_diffusion.monitor_locations
        return [g for g, e in locations.evaluate.items() if e]

    def prepare(self) -> ComputationalSpace:
        """Instantiates :class:`~.ComputationalSpace` instance.

        Generates a :class:`~.ComputationalSpace` instance from
        :attr:`settings`. 

        Returns
        -------
        :class:`~.ComputationalSpace`
            The Computational Space created from :attr:`settings`. 
        """
        print("Preparing Eddy Diffusion run...")
        restrict = {"models": "eddy_diffusion"}
        return ComputationalSpace(self.settings, restrict)
    
    def evaluate(self) -> None:
        """Loops over all elements in :attr:`space` and evaluates the model.

        Returns
        -------
        None

        """
        for setting in self.space.space:
            count = f"{self.space.linear_index(setting) + 1}/{len(self.space)}"
            print(f"Evaluating computational space element {count}")
            self.run(setting)

    def run(self, setting: RIDTConfig) -> None:
        """Evaluates the model for a set of parameters, for all geometries.
    
        Loops over all monitor locations that have been selected for evaluation
        and evaluates them over their respective domains. Writes output to
        :attr:`data_store`.

        Parameters
        ----------
        settings : :class:`~.RIDTConfig`
            The settings for the run in question.

        Returns
        -------
        None

        """
        self.data_store.add_run(setting)

        domain = Domain(setting)
        solver = EddyDiffusion(setting)
        locations = setting.models.eddy_diffusion.monitor_locations

        for geometry in self.geometries:
            print(f"Evaluating {geometry} monitor locations...")
            for name, item in getattr(locations, geometry).items():
                print(f"Evaluating {name}...")
                space = getattr(domain, geometry)(item)
                output = solver(*getattr(domain, geometry)(item), domain.time)
                self.data_store[setting].add(geometry, name, squeeze(output))
    
    def compute_exposure(self) -> None:
        """Computes the exposure from the concentration data.

        Saves the result to a new data store :attr:`exposure_store`

        Returns
        -------
        None

        """
        if self.settings.compute_exposure:
            print("\nComputing exposure...")
            self.exposure_store = Exposure(self.settings, self.data_store)

    def args(self, data_store: BatchDataStore, quantity: str) -> tuple:
        """A helper function that returns a tuple of some values.

        Parameters
        ----------

        data_store : :class:`~.BatchDataStore`
            A batch data store.

        quantity : :obj:`str`
            The name of the quantity in the data store.

        Returns
        -------
        :obj:`tuple`
            The various quantites as a tuple.

        """
        return self.settings, data_store, self.space, self.outdir, quantity

    def write(self) -> None:
        """Writes all data stores to disk.

        Returns
        -------
        None

        """
        print("\nWriting data to disk... ")
        BatchDataStoreWriter(*self.args(self.data_store, "concentration"))
        if self.settings.compute_exposure:
            BatchDataStoreWriter(*self.args(self.exposure_store, "exposure"))

    def plot(self) -> None:
        """Plots all requested data and writes it to disk.

        Returns
        -------
        None

        """
        print("\nProducing plots... ")
        BatchDataStorePlotter(*self.args(self.data_store, "concentration"))
        if self.settings.compute_exposure:
            BatchDataStorePlotter(*self.args(self.exposure_store, "exposure"))

    def analyse(self) -> None:
        """Performs all relevant analysis and writes it to disk.

        Returns
        -------
        None

        """
        if not self.settings.models.eddy_diffusion.analysis.perform_analysis:
            return
        print("\nPerforming data analysis...")
        BatchDataStoreAnalyser(*self.args(self.data_store, "concentration"))
        if self.settings.compute_exposure:
            BatchDataStoreAnalyser(*self.args(self.exposure_store, "exposure"))




import unittest
import os
import shutil
from pathlib import Path

from ridt.base import ComputationalSpace

from ridt.config import ConfigFileParser

from ridt.data.batchdatastore import BatchDataStoreIDError
from ridt.data import BatchDataStore
from ridt.data import BatchDataStoreWriter


class ST12(unittest.TestCase):

    """System Test 12. Test the system
       can write JSON files to disk."""

    def setUp(self) -> None:

        """setUp method that instantiates the
           :class:`~.RIDTConfig` class and the
           :class:`~.ComputationalSpace` class."""

        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.directory = os.path.join(this_dir, "st12")
        Path(self.directory).mkdir(parents=True, exist_ok=True)


        with ConfigFileParser(os.path.join(this_dir, "st11/config.json")) as cfp:
            self.c = cfp

        restrict = {"models": "well_mixed"}
        self.space = ComputationalSpace(self.c, restrict)

    def tearDown(self) -> None:
        shutil.rmtree(self.directory)

    def test_verify(self):

        """Verifies that the config.json file is created
        in the correct directory, then removes the created
        files."""

        bds = BatchDataStore()
        for space in self.space.space:
            bds.add_run(space)

        try:
            BatchDataStoreWriter(
                self.c, bds, self.space, self.directory, "concentration")
        except BatchDataStoreIDError as e:
            pass
        self.assertIn(
            "batch_config.json", os.listdir(f"{self.directory}")
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import shutil

from os.path import join
from os.path import dirname
from os.path import isfile
from os.path import isdir
from os.path import islink
from os import listdir
from os import system
from os import unlink

from pathlib import Path

from numpy import genfromtxt
from numpy import load
from numpy import delete
from numpy import ndarray
from numpy import array
from numpy import abs

from ridt.config import ConfigFileParser

from ridt.config.ridtconfig import FixedDurationSource
from ridt.config.ridtconfig import InstantaneousSource
from ridt.config.ridtconfig import InfiniteDurationSource


def delete_folder_contents(path: str):
    for filename in [f for f in listdir(path) if not f.endswith(".gitkeep")]:
        file_path = join(path, filename)
        try:
            if isfile(file_path) or islink(file_path):
                unlink(file_path)
            elif isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


class ST28(unittest.TestCase):
    """System Test 28. Check that the eddy diffusion model output matches the 
    verification data."""

    def conf(self, cat: str):
        return join(self.root, cat, "config.json")
    
    def out(self, cat: str):
        return join(self.root, cat, "out")
    
    def points(self, cat: str):
        path = join(self.root, cat, "points.csv")
        with open(path, 'r', encoding='utf-8-sig') as f:
            return genfromtxt(f, delimiter=',')
    
    def lines(self, cat: str):
        path = join(self.root, cat, "lines.csv")
        with open(path, 'r', encoding='utf-8-sig') as f:
            return genfromtxt(f, delimiter=',')
    
    def percentage_error(self, test_data: ndarray, val_data: ndarray):
        return abs(((test_data - val_data) / val_data)) * 100

    def loadpoint(self, cat: str, num: int):
        path = join(
            self.root,
            cat,
            "out",
            "points",
            "concentration",
            "data",
            f"point_{num}.npy")
        return load(path)

    def loadline(self, cat: str, num: int):
        path = join(
            self.root,
            cat,
            "out",
            "lines",
            "concentration",
            "data",
            f"line_{num}.npy")
        return load(path)

    def setUp(self) -> None:
        self.root = join(dirname(__file__), "st28")
        self.cat = [
            "insacr0point",
            "insacr0line",
            "insacr5point",
            "insacr5line",
            "infacr0point",
            "infacr0line",
            "infacr5point",
            "infacr5line",
            "fixacr0point",
            "fixacr0line",
            "fixacr5point"
            "fixacr5line"
        ]

    def test_insacr0point(self):
        idstr = "insacr0point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")
        print(self.out(idstr))

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))

    
    def test_insacr0line(self):
        idstr = "insacr0line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))
   
    
    def test_insacr5point(self):
        idstr = "insacr5point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))

    
    def test_insacr5line(self):
        idstr = "insacr5line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))


    
    def test_infacr5point(self):
        idstr = "infacr5point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))

    
    def test_infacr5line(self):
        idstr = "infacr5line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        delete_folder_contents(self.out(idstr))
    
    
    def test_infacr0point(self):
        idstr = "infacr0point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))

    
    def test_infacr0line(self):
        idstr = "infacr0line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        delete_folder_contents(self.out(idstr))

    
    def test_fixacr5point(self):
        idstr = "fixacr5point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))
    
    def test_fixacr5line(self):
        idstr = "fixacr5line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        delete_folder_contents(self.out(idstr))
    
    def test_fixacr0point(self):
        idstr = "fixacr0point"
        print("")
        print(idstr)
        valpoints = self.points(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        points = [self.loadpoint(idstr, num) for num in range(1, 4)]
        points = [array([p for i, p in enumerate(point)]) for point in points]
        points = [delete(point, 0) for point in points]

        for idp, point in enumerate(points):
            for idx, item in enumerate(self.percentage_error(point, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        print("")
        delete_folder_contents(self.out(idstr))

    def test_fixacr0line(self):
        idstr = "fixacr0line"
        print("")
        print(idstr)
        valpoints = self.lines(idstr)

        system(f"ridt run {self.conf(idstr)} {self.out(idstr)}")

        lines = [self.loadline(idstr, num) for num in range(1, 4)]
        lines = [line[-1, :] for line in lines]

        for idp, line in enumerate(lines):
            for idx, item in enumerate(self.percentage_error(line, valpoints[:, idp])):
                self.assertTrue(item < 0.1)
        delete_folder_contents(self.out(idstr))



if __name__ == "__main__":
    unittest.main()


from typing import Tuple

from itertools import product

from numpy import linspace
from numpy import meshgrid
from numpy import unravel_index

from ridt.config.ridtconfig import RIDTConfig
from ridt.config.ridtconfig import Point
from ridt.config.ridtconfig import Line 
from ridt.config.ridtconfig import Plane 


class Domain:
    """A class the provides various spatial and temporal domain information.

    This class provides various functions that give domain infomation relating
    to monitor locations and the total domain. The meshgrids over which the
    equations are evaluated are generated from the settings object by this
    class.
    
    Attributes
    ----------
    set : :obj:`~.RIDTConfig`
        The settings object for the current run.
    
    """

    def __init__(self, settings: RIDTConfig):
        """[summary]

        Parameters
        ----------
        setting : :class:`~.RIDTConfig`
            The settings object containing the units selections and definitions.

        """
        self.set = settings
        self._x = linspace(0.0, self.set.dimensions.x, self.set.spatial_samples.x)
        self._y = linspace(0.0, self.set.dimensions.y, self.set.spatial_samples.y)
        self._z = linspace(0.0, self.set.dimensions.z, self.set.spatial_samples.z)
        self._time = linspace(0.0, self.set.total_time, self.set.time_samples)

    @property
    def x(self):
        """:obj:`Iterable`[:obj:`float`] : The discretised x domain."""
        return self._x

    @property
    def y(self):
        """:obj:`Iterable`[:obj:`float`] : The discretised y domain."""
        return self._y
    
    @property
    def z(self):
        """:obj:`Iterable`[:obj:`float`] : The discretised z domain."""
        return self._z
   
    @property
    def full(self):
        """:obj:`Tuple`[:class:`~numpy.ndarray`] : The full meshgrid domain."""
        return meshgrid(self.x, self.y, self.z, indexing='ij')
    
    @property
    def time(self):
        """:obj:`Iterable`[:obj:`float`] : The discretised time domain."""
        return self._time
    
    def point_cartesian(self, point: Point):
        """Returns the values for the point's position as a tuple

        Parameters
        ----------
        point : :class:`~.Point`
            The point being queried.

        Returns
        -------
        :obj:`tuple`[:obj:`float`]
            The (x, y, z) position of the point.
            
        """
        return (point.x, point.y, point.z)
    
    def line_cartesian(self, line: Line):
        """Returns a list of tuples with all poistions on the line and
        their corresponding indices.

        Parameters
        ----------
        line : :class:`~.Line`
            The line being queried.

        Returns
        -------
        :obj:`list`[:obj:`tuple`[:obj:`tuple`[:obj:`int`], :obj:`tuple`[:obj:`float`]]]
            The grid indices and values of all points in the line.

        """
        if line.parallel_axis == "x":
            return [
                (unravel_index(i, self.x.shape), item) for i, item in 
                enumerate(product(self.x, [line.point.y], [line.point.z]))
            ]
        elif line.parallel_axis == "y":
            return [
                (unravel_index(i, self.y.shape), item) for i, item in 
                enumerate(product([line.point.x], self.y, [line.point.z]))
            ]
        else:
            return [
                (unravel_index(i, self.z.shape), item) for i, item in 
                enumerate(product([line.point.x], [line.point.y], self.z))
            ]
    
    def plane_cartesian(self, plane: Plane):
        """Returns a list of tuples with all poistions on the plane and
        their corresponding indices.

        Parameters
        ----------
        plane : :class:`~.Plane`
            The plane being queried.

        Returns
        -------
        :obj:`list`[:obj:`tuple`[:obj:`tuple`[:obj:`int`], :obj:`tuple`[:obj:`float`]]]
            The grid indices and values of all points in the plane.

        """
        if plane.axis == "xy":
            return [
                (unravel_index(i, (len(self.x), len(self.y))), item) for i, item in 
                enumerate(product(self.x, self.y, [plane.distance]))
            ]
        elif plane.axis == "yz":
            return [
                (unravel_index(i, (len(self.y), len(self.z))), item) for i, item in 
                enumerate(product([plane.distance], self.y, self.z))
            ]
        else:
            return [
                (unravel_index(i, (len(self.x), len(self.z))), item) for i, item in 
                enumerate(product(self.x, [plane.distance], self.z))
            ]
    
    def domain_cartesian(self, *args, **kwargs):
        """Returns a list of tuples with all poistions on the domain and
        their corresponding indices.

        Returns
        -------
        :obj:`list`[:obj:`tuple`[:obj:`tuple`[:obj:`int`], :obj:`tuple`[:obj:`float`]]]
            The grid indices and values of all points in the domain.

        """
        return [
            (unravel_index(i, (len(self.x), len(self.y), len(self.z))), item)
            for i, item in enumerate(product(self.x, self.y, self.z))
        ]
    
    def points(self, point: Point):
        """Returns the meshgrid corresponding to a point-like monitor locaiton.

        Parameters
        ----------
        point : :class:`~.Point`
            The point-line monitor location settings object.

        Returns
        -------
        :obj:`Tuple`[:class:`~numpy.ndarray`]
            The 3D meshgrid for the point.
        
        """
        return meshgrid(point.x, point.y, point.z, indexing="ij")

    def lines(self, line: Line):
        """Returns the meshgrid corresponding to a line-like monitor locaiton.

        Parameters
        ----------
        line : :class:`~.Line`
            The line-line monitor location settings object.

        Returns
        -------
        :obj:`Tuple`[:class:`~numpy.ndarray`]
            The 3D meshgrid for the line.
        
        """
        if line.parallel_axis == "x":
            return meshgrid(self.x, line.point.y, line.point.z, indexing="ij")
        elif line.parallel_axis == "y":
            return meshgrid(line.point.x, self.y, line.point.z, indexing="ij")
        else:
            return meshgrid(line.point.x, line.point.y, self.z, indexing="ij")

    def planes(self, plane: Plane):
        """Returns the meshgrid corresponding to a plane-like monitor locaiton.

        Parameters
        ----------
        plane : :class:`~.Plane`
            The plane-plane monitor location settings object.

        Returns
        -------
        :obj:`Tuple`[:class:`~numpy.ndarray`]
            The 3D meshgrid for the plane.
        
        """
        if plane.axis == "xy":
            rv = meshgrid(self.x, self.y, plane.distance, indexing="ij")
            return rv
        elif plane.axis == "yz":
            return  meshgrid(plane.distance, self.y, self.z, indexing="ij")
        else:
            return meshgrid(self.x, plane.distance, self.z, indexing="ij")
    
    def domain(self, *args, **kwargs):
        """An interface for the :attr:`full` parameter.
        
        Returns
        -------
        :obj:`Tuple`[:class:`~numpy.ndarray`]
            The 3D meshgrid for the whole domain.
        
        """
        return self.full
    
    def values(self, id: str, index: Tuple[int]):
        """Returns the spatio-temporal values for an index and monitor location

        This method takes a monitor location and some temporal meshgrid indices 
        and returns the corresponding spatial values.

        Parameters
        ----------
        id : :obj:`str`
            The monitor location id string.

        index : :obj:`Tuple`[:obj:`int`]
            The temporal meshgrid index being queried.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        Raises
        ------
        :obj:`ValueError`
            If an index with an invalid number of dimensions is passed.

        """
        if id == "well_mixed":
            return self.wellmixed_values(index)
        elif len(index) == 1:
            return self.point_values(id, index) 
        elif len(index) == 2:
            return self.line_values(id, index) 
        elif len(index) == 3:
            return self.plane_values(id, index) 
        elif len(index) == 4:
            return self.domain_values(id, index) 
        else:
            raise ValueError("Not a valid number of dimensions.")
        
    def wellmixed_values(self, index: Tuple[int]):
        """Returns the spatio-temporal values for an index, for well mixed model

        Parameters
        ----------
        index : :obj:`Tuple`[:obj:`int`]
            The index being queried.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        """
        return self.time[index[0]], "N/A", "N/A", "N/A" 

    def point_values(self, id: str, index: Tuple[int]):
        """Returns the spatio-temporal values for an index, for points. 

        Parameters
        ----------
        index : :obj:`Tuple`[:obj:`int`]
            The index being queried.

        id : :obj:`str`
            The point id string.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        """
        point = self.set.models.eddy_diffusion.monitor_locations.points[id]
        return self.time[index[0]], point.x, point.y, point.z
        
    def line_values(self, id: str, index: Tuple[int]):
        """Returns the spatio-temporal values for an index, for lines. 

        Parameters
        ----------
        index : :obj:`Tuple`[:obj:`int`]
            The index being queried.

        id : :obj:`str`
            The line id string.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        """
        line = self.set.models.eddy_diffusion.monitor_locations.lines[id]
        if line.parallel_axis == "x":
            return self.time[index[0]], self.x[index[1]], line.point.y, line.point.z
        elif line.parallel_axis == "y":
            return self.time[index[0]], line.point.x, self.y[index[1]], line.point.z
        else:
            return self.time[index[0]], line.point.x, line.point.y, self.z[index[1]] 
        
    def plane_values(self, id: str, index: Tuple[int]):
        """Returns the spatio-temporal values for an index, for planes. 

        Parameters
        ----------
        index : :obj:`Tuple`[:obj:`int`]
            The index being queried.

        id : :obj:`str`
            The plane id string.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        """
        plane = self.set.models.eddy_diffusion.monitor_locations.planes[id]
        if plane.axis == "xy":
            return self.time[index[0]], self.x[index[1]], self.y[index[2]], plane.distance
        elif plane.axis == "yz":
            return self.time[index[0]], plane.distance, self.y[index[1]], self.z[index[2]]
        else:
            return self.time[index[0]], self.x[index[1]], plane.distance, self.z[index[2]] 
    
    def domain_values(self, id: str, index: Tuple[int]):
        """Returns the spatio-temporal values for an index, for the domain. 

        Parameters
        ----------
        index : :obj:`Tuple`[:obj:`int`]
            The index being queried.

        id : :obj:`str`
            The domain id string.

        Returns
        -------
        :obj:`Tuple`[:obj:`float`]
            The time, x position, y position, and z position.

        """
        return self.time[index[0]], self.x[index[1]], self.y[index[2]], self.z[index[3]] 